from ..utils.nointerrupt import WithKeyboardInterruptAs
from .state import Concretize, TerminateState
import logging
import multiprocessing
import threading
import os
import typing


logger = logging.getLogger(__name__)
# logger.setLevel(9)


# Workers
# There are 4 types of Workers
# WorkerSingle: run over main process and will not provide any concurrency
# WorkerThread: runs on a different thread
# WorkerProcess: runs on a different process - Full multiprocessing
# WorkerMultiprocessing: --planned-- runs on a different computer


class Worker:
    """
    A Manticore Worker.
    This will run forever potentially in a different process. Normally it
    will be spawned at Manticore constructor and will stay alive until killed.
    A Worker can be in 3 phases: STANDBY, RUNNING, KILLED. And will react to
    different events: start, stop, kill.
    The events are transmitted via 2 conditional variable: m._killed and
    m._started.

    .. code-block:: none

        STANDBY:   Waiting for the start event
        RUNNING:   Exploring and spawning states until no more READY states or
        the cancel event is received
        KIlLED:    This is the end. No more manticoring in this worker process

                     +---------+     +---------+
                +--->+ STANDBY +<--->+ RUNNING |
                     +-+-------+     +-------+-+
                       |                     |
                       |      +--------+     |
                       +----->+ KILLED <-----+
                              +----+---+
                                   |
                                   #
    """

    def __init__(self, *, id, manticore, single=False):
        self.manticore = manticore
        self.id = id
        self.single = single

    def start(self):
        raise NotImplementedError

    def join(self):
        raise NotImplementedError

    def run(self, *args):
        # This controls the main symbolic execution loop of one of the workers
        logger.debug(
            "Starting Manticore Symbolic Emulator Worker %d. Pid %d Tid %d).",
            self.id,
            os.getpid(),
            threading.get_ident(),
        )

        m = self.manticore
        current_state = None
        m._publish("will_start_worker", self.id)

        # If CTRL+C is received at any worker lets abort exploration via m.kill()
        # kill will set m._killed flag to true and then each worker will slowly
        # get out of its mainloop and quit.
        with WithKeyboardInterruptAs(m.kill):

            # The worker runs until the manticore is killed
            while not m._killed.value:

                # STARTED - Will try to consume states until a STOP event is received
                # Outer loop, Keep getting states until someone request us to STOP
                try:  # handle fatal errors even exceptions in the exception handlers
                    try:  # handle Concretize and TerminateState

                        # At RUNNING
                        # The START has been requested, we operate with under the assumption
                        # that manticore we will let us stay at this phase for a _while_
                        # Requests to STOP will be honored ASAP (i.e. Not immediately)

                        # Select a single state
                        # wait for other worker to add states to the READY list
                        # This momentarily get the main lock and then releases
                        # it while waiting for changes
                        # Raises an Exception if manticore gets cancelled
                        # while waiting or if there are no more potential states
                        logger.debug("[%r] Waiting for states", self.id)
                        # If at STANDBY wait for any change
                        current_state = m._get_state(wait=True)

                        # there are no more states to process
                        # states can come from the ready list or by forking
                        # states currently being analyzed in the busy list
                        if current_state is None:
                            logger.debug("[%r] No more states", self.id)
                            break

                        # assert current_state is not None
                        # Allows to terminate manticore worker on user request
                        # even in the middle of an execution
                        logger.debug("[%r] Running", self.id)
                        assert (
                            current_state.id in m._busy_states
                            and current_state.id not in m._ready_states
                        )

                        # This does not hold the lock so we may loss some event
                        # flickering
                        while not m._killed.value:
                            current_state.execute()
                        else:
                            logger.debug("[%r] Stopped and/or Killed", self.id)
                            # On going execution was stopped or killed. Lets
                            # save any progress on the current state using the
                            # same id. No other worker will use this state in
                            # this run
                            m._save(current_state, state_id=current_state.id)
                            m._revive_state(current_state.id)
                            current_state = None

                        assert current_state is None
                    # Handling Forking and terminating exceptions
                    except Concretize as exc:
                        logger.debug("[%r] Performing %r", self.id, exc.message)
                        # The fork() method can decides which state to keep
                        # exploring. For example when the fork results in a
                        # single state it is better to just keep going.
                        # Though, normally fork() saves the spawned childs,
                        # returns a None and let _get_state choose what to explore
                        # next
                        m._fork(current_state, exc.expression, exc.policy, exc.setstate)
                        current_state = None

                    except TerminateState as exc:
                        logger.debug("[%r] Debug State %r %r", self.id, current_state, exc)
                        # Notify this state is done
                        m._publish("will_terminate_state", current_state, exc)
                        # Update the stored version of the current state

                        current_state._terminated_by = exc

                        m._save(current_state, state_id=current_state.id)
                        # Add the state to the terminated state list re-using
                        # the same id. No other worker will use this state in
                        # this run
                        m._terminate_state(current_state.id)

                        m._publish("did_terminate_state", current_state, exc)
                        current_state = None

                except (Exception, AssertionError) as exc:
                    import traceback

                    formatted = traceback.format_exc()
                    logger.error("Exception in state %r: %r\n%s ", self.id, exc, formatted)
                    # Internal Exception
                    # Add the state to the terminated state list
                    if current_state is not None:
                        # Drop any work on this state in case it is inconsistent

                        # Update the stored version of the current state
                        # Saved to a fresh id in case other worker have an old
                        # version this state cached over the old id
                        m._publish("will_kill_state", current_state, exc)
                        m._save(current_state, state_id=current_state.id)
                        m._kill_state(current_state.id)
                        m._publish("did_kill_state", current_state, exc)
                        current_state = None
                    break

            # Getting out.
            # At KILLED
            logger.debug("[%r] Getting out of the mainloop", self.id)
            m._publish("did_terminate_worker", self.id)


class WorkerSingle(Worker):
    """A single worker that will run in the current process and current thread.
    As this will not provide any concurrency is normally only used for
    profiling underlying arch emulation and debugging."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, single=True, **kwargs)

    def start(self):
        self.run()

    def join(self):
        pass


class WorkerThread(Worker):
    """ A worker thread """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._t = None

    def start(self):
        self._t = threading.Thread(target=self.run)
        self._t.start()

    def join(self):
        self._t.join()
        self._t = None


class WorkerProcess(Worker):
    """ A worker process """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._p = None

    def start(self):
        self._p = multiprocessing.Process(target=self.run)
        self._p.start()

    def join(self):
        self._p.join()
        self._p = None


class DaemonThread(WorkerThread):
    """
    Special case of WorkerThread that will exit whenever the main Manticore process exits.
    """

    def start(self, target: typing.Optional[typing.Callable] = None):
        """
        Function that starts the thread. Can take an optional callable to be invoked at the start, or can be subclassed,
        in which case `target` should be None and the the `run` method will be invoked at the start.

        :param target: an optional callable that will be invoked to start the thread. The callable should accept this
        thread as an argument.
        """
        logger.debug(
            "Starting Daemon %d. (Pid %d Tid %d).",
            self.id,
            os.getpid(),
            threading.get_ident(),
        )

        self._t = threading.Thread(target=self.run if target is None else target, args=(self,))
        self._t.daemon = True
        self._t.start()



import logging

from functools import wraps
from typing import Any, Callable, TypeVar

from ..utils.event import Eventful


logger = logging.getLogger(__name__)


class OSException(Exception):
    pass


T = TypeVar("T")


def unimplemented(wrapped: Callable[..., T]) -> Callable[..., T]:
    @wraps(wrapped)
    def new_wrapped(self: Any, *args, **kwargs) -> T:
        cpu = getattr(getattr(self, "parent", None), "current", None)
        pc_str = "<unknown PC>" if cpu is None else hex(cpu.read_register("PC"))
        logger.warning(
            f"Unimplemented system call: %s: %s(%s)",
            pc_str,
            wrapped.__name__,
            ", ".join(hex(a) if isinstance(a, int) else str(a) for a in args),
        )
        return wrapped(self, *args, **kwargs)

    return new_wrapped


class SyscallNotImplemented(OSException):
    """
    Exception raised when you try to call an unimplemented system call.
    Go to linux.py and add an implementation!
    """

    def __init__(self, idx, name):
        msg = f'Syscall index "{idx}" ({name}) not implemented.'
        super().__init__(msg)


class Platform(Eventful):
    """
    Base class for all platforms e.g. operating systems or virtual machines.
    """

    _published_events = {"solve"}

    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)

    def invoke_model(self, model, prefix_args=None):
        self._function_abi.invoke(model, prefix_args)

    def __setstate__(self, state):
        super().__setstate__(state)

    def __getstate__(self):
        state = super().__getstate__()
        return state

    def generate_workspace_files(self):
        return {}



###############################################################################
# Solver
# A solver maintains a companion smtlib capable process connected via stdio.
# It can be in 4 main states: None, unknown, sat, unsat
# None      nothing was yet sent to the smtlib process. Al the state is only at the python side
# unknown   is an error state. Some query sent earlier was unsuccessful or timed out.
#           Further interaction with the smtlib process will probably keep returning
#           unknown. An exception is raised.
# sat       the current set of constraints is satisfiable and has at least one solution
# unsat     the current set of constraints is impossible
#
# It starts at None.
# Once you Solver.check() it the status is changed to sat or unsat (or unknown+exception)
# You can create new symbols operate on them. The declarations will be sent to the smtlib process when needed.
# You can add new constraints. A new constraint may change the state from {None, sat} to {sat, unsat, unknown}

import os
import shutil
import threading
from queue import Queue
import collections
import shlex
import time
from functools import lru_cache
from typing import Dict, Tuple, Sequence, Optional
from subprocess import PIPE, Popen, check_output
import re
from . import operators as Operators
from .constraints import *
from .visitors import *
from ...exceptions import Z3NotFoundError, SolverError, SolverUnknown, TooManySolutions, SmtlibError
from ...utils import config
from . import issymbolic


class SolverType(config.ConfigEnum):
    """Used as configuration constant for choosing solver flavor"""

    z3 = "z3"
    cvc4 = "cvc4"
    yices = "yices"
    auto = "auto"


logger = logging.getLogger(__name__)
consts = config.get_group("smt")
consts.add("timeout", default=120, description="Timeout, in seconds, for each Z3 invocation")
consts.add("memory", default=1024 * 8, description="Max memory for Z3 to use (in Megabytes)")
consts.add(
    "maxsolutions",
    default=10000,
    description="Maximum solutions to provide when solving for all values",
)
consts.add("z3_bin", default="z3", description="Z3 solver binary to use")
consts.add("cvc4_bin", default="cvc4", description="CVC4 solver binary to use")
consts.add("yices_bin", default="yices-smt2", description="Yices solver binary to use")


consts.add("defaultunsat", default=True, description="Consider solver timeouts as unsat core")
consts.add(
    "optimize", default=True, description="Use smtlib command optimize to find min/max if available"
)

consts.add(
    "solver",
    default=SolverType.auto,
    description="Choose default smtlib2 solver (z3, yices, cvc4, auto)",
)

# Regular expressions used by the solver
RE_GET_EXPR_VALUE_FMT_BIN = re.compile(r"\(\((?P<expr>(.*))[ \n\s]*#b(?P<value>([0-1]*))\)\)")
RE_GET_EXPR_VALUE_FMT_DEC = re.compile(r"\(\((?P<expr>(.*))\ \(_\ bv(?P<value>(\d*))\ \d*\)\)\)")
RE_GET_EXPR_VALUE_FMT_HEX = re.compile(r"\(\((?P<expr>(.*))\ #x(?P<value>([0-9a-fA-F]*))\)\)")
RE_OBJECTIVES_EXPR_VALUE = re.compile(
    r"\(objectives.*\((?P<expr>.*) (?P<value>\d*)\).*\).*", re.MULTILINE | re.DOTALL
)
RE_MIN_MAX_OBJECTIVE_EXPR_VALUE = re.compile(r"(?P<expr>.*?)\s+\|->\s+(?P<value>.*)", re.DOTALL)


class SingletonMixin(object):
    __singleton_instances: Dict[Tuple[int, int], "SingletonMixin"] = {}

    @classmethod
    def instance(cls):
        tid = threading.get_ident()
        pid = os.getpid()
        if not (pid, tid) in cls.__singleton_instances:
            cls.__singleton_instances[(pid, tid)] = cls()
        return cls.__singleton_instances[(pid, tid)]


class SolverException(SmtlibError):
    """
    Solver exception
    """

    pass


class Solver(SingletonMixin):
    def __init__(self):
        pass

    def optimize(self, constraints, X, operation, M=10000):
        """
        Iteratively finds the maximum or minimal value for the operation
        (Normally Operators.UGT or Operators.ULT)

        :param constraints: the constraints set
        :param X: a symbol or expression
        :param M: maximum number of iterations allowed
        """
        raise SmtlibError("Abstract method not implemented")

    def check(self, constraints) -> bool:
        """Check if given constraints can be valid"""
        return self.can_be_true(constraints, True)

    def can_be_true(self, constraints, expression=True) -> bool:
        """Check if given expression could be valid"""
        raise SolverException("Abstract method not implemented")

    def must_be_true(self, constraints, expression) -> bool:
        """Check if expression is True and that it can not be False with current constraints"""
        solutions = self.get_all_values(constraints, expression, maxcnt=2, silent=True)
        return solutions == [True]

    def get_all_values(self, constraints, x, maxcnt=10000, silent=False):
        """Returns a list with all the possible values for the symbol x"""
        raise SolverException("Abstract method not implemented")

    def get_value(self, constraints, expression):
        """Ask the solver for one possible result of given expression using given set of constraints."""
        raise SolverException("Abstract method not implemented")

    def max(self, constraints, X: BitVec, M=10000):
        """
        Iteratively finds the maximum value for a symbol within given constraints.
        :param X: a symbol or expression
        :param M: maximum number of iterations allowed
        """
        assert isinstance(X, BitVec)
        return self.optimize(constraints, X, "maximize", M)

    def min(self, constraints, X: BitVec, M=10000):
        """
        Iteratively finds the minimum value for a symbol within given constraints.

        :param constraints: constraints that the expression must fulfil
        :param X: a symbol or expression
        :param M: maximum number of iterations allowed
        """
        assert isinstance(X, BitVec)
        return self.optimize(constraints, X, "minimize", M)

    def minmax(self, constraints, x, iters=10000):
        """Returns the min and max possible values for x within given constraints"""
        if issymbolic(x):
            m = self.min(constraints, x, iters)
            M = self.max(constraints, x, iters)
            return m, M
        else:
            return x, x


Version = collections.namedtuple("Version", "major minor patch")


class SmtlibProc:
    def __init__(self, command: str, debug: bool = False):
        """Single smtlib interactive process

        :param command: the shell command to execute
        :param debug: log all messaging
        """
        self._proc: Optional[Popen] = None
        self._command = command
        self._debug = debug

    def start(self):
        """Spawns POpen solver process"""
        if self._proc is not None:
            return
        self._proc = Popen(
            shlex.split(self._command),
            stdin=PIPE,
            stdout=PIPE,
            bufsize=0,
            universal_newlines=True,
            close_fds=True,
        )

    def stop(self):
        """
        Stops the solver process by:
        - sending a SIGKILL signal,
        - waiting till the process terminates (so we don't leave a zombie process)
        """
        if self._proc is None:
            return
        # if it did not finished already
        if self._proc.returncode is None:
            self._proc.stdin.close()
            self._proc.stdout.close()
            # Kill the process
            self._proc.kill()
            self._proc.wait()

            # No need to wait for termination, zombies avoided.
        self._proc = None

    def __readline_and_count(self):
        assert self._proc
        assert self._proc.stdout
        buf = self._proc.stdout.readline()  # No timeout enforced here
        # If debug is enabled check if the solver reports a syntax error
        # Error messages may contain an unbalanced parenthesis situation
        if self._debug:
            if "(error" in buf:
                raise SolverException(f"Error in smtlib: {buf}")
        lparen, rparen = map(sum, zip(*((c == "(", c == ")") for c in buf)))
        return buf, lparen, rparen

    def send(self, cmd: str) -> None:
        """
        Send a string to the solver.

        :param cmd: a SMTLIBv2 command (ex. (check-sat))
        """
        if self._debug:
            logger.debug(">%s", cmd)
        self._proc.stdout.flush()  # type: ignore
        self._proc.stdin.write(f"{cmd}\n")  # type: ignore

    def recv(self) -> str:
        """Reads the response from the smtlib solver"""
        buf, left, right = self.__readline_and_count()
        bufl = [buf]
        while left != right:
            buf, l, r = self.__readline_and_count()
            bufl.append(buf)
            left += l
            right += r

        buf = "".join(bufl).strip()

        if self._debug:
            logger.debug("<%s", buf)

        return buf

    def _restart(self) -> None:
        """Auxiliary to start or restart the external solver"""
        self.stop()
        self.start()

    def is_started(self):
        return self._proc is not None


class SMTLIBSolver(Solver):
    def __init__(
        self,
        command: str,
        init: Sequence[str] = None,
        value_fmt: int = 16,
        support_reset: bool = False,
        support_minmax: bool = False,
        support_pushpop: bool = False,
        multiple_check: bool = True,
        debug: bool = False,
    ):

        """
        Build a smtlib solver instance.
        This is implemented using an external solver (via a subprocess).
        """
        super().__init__()
        self._smtlib: SmtlibProc = SmtlibProc(command, debug)

        # Commands used to initialize smtlib
        if init is None:
            init = tuple()
        self._init = init
        self._get_value_fmt = (
            {
                2: RE_GET_EXPR_VALUE_FMT_BIN,
                10: RE_GET_EXPR_VALUE_FMT_DEC,
                16: RE_GET_EXPR_VALUE_FMT_HEX,
            }[value_fmt],
            value_fmt,
        )

        self._support_minmax = support_minmax
        self._support_reset = support_reset
        self._support_pushpop = support_pushpop
        self._multiple_check = multiple_check

        if not self._support_pushpop:
            setattr(self, "_push", None)
            setattr(self, "_pop", None)

        if self._support_minmax and consts.optimize:
            setattr(self, "optimize", self._optimize_fancy)
        else:
            setattr(self, "optimize", self._optimize_generic)

        self._smtlib.start()
        # run solver specific initializations

    def _reset(self, constraints: Optional[str] = None) -> None:
        """Auxiliary method to reset the smtlib external solver to initial defaults"""
        if self._support_reset:
            self._smtlib.start()  # does not do anything if already started
            self._smtlib.send("(reset)")
        else:
            self._smtlib.stop()  # does not do anything if already stopped
            self._smtlib.start()

        for cfg in self._init:
            self._smtlib.send(cfg)

        if constraints is not None:
            self._smtlib.send(constraints)

    # UTILS: check-sat get-value
    def _is_sat(self) -> bool:
        """
        Check the satisfiability of the current state

        :return: whether current state is satisfiable or not.
        """
        start = time.time()
        self._smtlib.send("(check-sat)")
        status = self._smtlib.recv()
        logger.debug("Check took %s seconds (%s)", time.time() - start, status)
        if status not in ("sat", "unsat", "unknown"):
            raise SolverError(status)
        if consts.defaultunsat:
            if status == "unknown":
                logger.info("Found an unknown core, probably a solver timeout")
                status = "unsat"
        if status == "unknown":
            raise SolverUnknown(status)
        return status == "sat"

    def _assert(self, expression: Bool):
        """Auxiliary method to send an assert"""
        smtlib = translate_to_smtlib(expression)
        self._smtlib.send(f"(assert {smtlib})")

    def __getvalue_bv(self, expression_str: str) -> int:
        self._smtlib.send(f"(get-value ({expression_str}))")
        pattern, base = self._get_value_fmt
        m = pattern.match(self._smtlib.recv())
        expr, value = m.group("expr"), m.group("value")  # type: ignore
        return int(value, base)

    def __getvalue_bool(self, expression_str):
        self._smtlib.send(f"(get-value ({expression_str}))")
        ret = self._smtlib.recv()
        return {"true": True, "false": False}[ret[2:-2].split(" ")[1]]

    def _getvalue(self, expression) -> Union[int, bool, bytes]:
        """
        Ask the solver for one possible assignment for given expression using current set of constraints.
        The current set of expressions must be sat.

        NOTE: This is an internal method: it uses the current solver state (set of constraints!).
        """
        if not issymbolic(expression):
            return expression

        if isinstance(expression, Array):
            result = bytearray()
            for c in expression:
                expression_str = translate_to_smtlib(c)
                result.append(self.__getvalue_bv(expression_str))
            return bytes(result)
        else:
            if isinstance(expression, BoolVariable):
                return self.__getvalue_bool(expression.name)
            elif isinstance(expression, BitVecVariable):
                return self.__getvalue_bv(expression.name)

        raise NotImplementedError(
            f"_getvalue only implemented for Bool, BitVec and Array. Got {type(expression)}"
        )

    # push pop
    def _push(self):
        """Pushes and save the current constraint store and state."""
        self._smtlib.send("(push 1)")

    def _pop(self):
        """Recall the last pushed constraint store and state."""
        self._smtlib.send("(pop 1)")

    @lru_cache(maxsize=32)
    def can_be_true(self, constraints: ConstraintSet, expression: Union[bool, Bool] = True) -> bool:
        """Check if two potentially symbolic values can be equal"""
        if isinstance(expression, bool):
            if not expression:
                return expression
            else:
                # if True check if constraints are feasible
                self._reset(constraints.to_string())
                return self._is_sat()

        with constraints as temp_cs:
            temp_cs.add(expression)
            self._reset(temp_cs.to_string())
            return self._is_sat()

    # get-all-values min max minmax
    def _optimize_generic(self, constraints: ConstraintSet, x: BitVec, goal: str, max_iter=10000):
        """
        Iteratively finds the maximum or minimum value for the operation
        (Normally Operators.UGT or Operators.ULT)

        :param constraints: constraints to take into account
        :param x: a symbol or expression
        :param goal: goal to achieve, either 'maximize' or 'minimize'
        :param max_iter: maximum number of iterations allowed
        """
        # TODO: consider adding a mode to return best known value on timeout
        assert goal in ("maximize", "minimize")
        operation = {"maximize": Operators.UGE, "minimize": Operators.ULE}[goal]

        last_value: Optional[Union[int, bool, bytes]] = None

        start = time.time()
        with constraints as temp_cs:
            X = temp_cs.new_bitvec(x.size)  # _getvalue needs a Variable
            temp_cs.add(X == x)
            self._reset(temp_cs.to_string())

            # Find one value and use it as currently known min/Max
            if not self._is_sat():
                raise SolverException("UNSAT")
            last_value = self._getvalue(X)
            self._assert(operation(X, last_value))

            # This uses a binary search to find a suitable range for aux
            # Use known solution as min or max depending on the goal
            if goal == "maximize":
                m, M = last_value, (1 << X.size) - 1
            else:
                m, M = 0, last_value

            # Iteratively divide the range
            L = None
            while L not in (M, m):
                L = (m + M) // 2
                self._assert(operation(X, L))
                sat = self._is_sat()

                # depending on the goal move one of the extremes
                if goal == "maximize" and sat or goal == "minimize" and not sat:
                    m = L
                else:
                    M = L

                if time.time() - start > consts.timeout:
                    raise SolverError("Timeout")

        # reset to before the dichotomic search
        with constraints as temp_cs:
            X = temp_cs.new_bitvec(x.size)  # _getvalue needs a Variable
            temp_cs.add(X == x)
            self._reset(temp_cs.to_string())

            # At this point we know aux is inside [m,M]
            # Lets constrain it to that range
            self._assert(Operators.UGE(X, m))
            self._assert(Operators.ULE(X, M))

            # And now check all remaining possible extremes
            last_value = None
            i = 0
            while self._is_sat():
                last_value = self._getvalue(X)
                self._assert(operation(X, last_value))
                self._assert(X != last_value)
                i = i + 1
                if i > max_iter:
                    raise SolverError("Optimizing error, maximum number of iterations was reached")
                if time.time() - start > consts.timeout:
                    raise SolverError("Timeout")
            if last_value is not None:
                return last_value
            raise SolverError("Optimizing error, unsat or unknown core")

    @lru_cache(maxsize=32)
    def get_all_values(
        self,
        constraints: ConstraintSet,
        expression,
        maxcnt: Optional[int] = None,
        silent: bool = False,
    ):
        """Returns a list with all the possible values for the symbol x"""
        if not isinstance(expression, Expression):
            return [expression]
        assert isinstance(expression, Expression)
        expression = simplify(expression)
        if maxcnt is None:
            maxcnt = consts.maxsolutions
            if isinstance(expression, Bool) and consts.maxsolutions > 1:
                # We know there is max 2 solutions when Bool
                maxcnt = 2
                silent = True

        with constraints as temp_cs:
            if isinstance(expression, Bool):
                var = temp_cs.new_bool()
            elif isinstance(expression, BitVec):
                var = temp_cs.new_bitvec(expression.size)
            elif isinstance(expression, Array):
                var = temp_cs.new_array(
                    index_max=expression.index_max,
                    value_bits=expression.value_bits,
                    taint=expression.taint,
                ).array
            else:
                raise NotImplementedError(
                    f"get_all_values only implemented for {type(expression)} expression type."
                )

            temp_cs.add(var == expression)
            self._reset(temp_cs.to_string())
            result = []
            start = time.time()
            while self._is_sat():
                value = self._getvalue(var)
                result.append(value)

                if len(result) >= maxcnt:
                    if silent:
                        # do not throw an exception if set to silent
                        # Default is not silent, assume user knows
                        # what they are doing and will check the size
                        # of returned vals list (previous smtlib behavior)
                        break
                    else:
                        raise TooManySolutions(result)
                if time.time() - start > consts.timeout:
                    if silent:
                        logger.info("Timeout searching for all solutions")
                        return list(result)
                    raise SolverError("Timeout")
                # Sometimes adding a new contraint after a check-sat eats all the mem
                if self._multiple_check:
                    self._smtlib.send(f"(assert {translate_to_smtlib(var != value)})")
                else:
                    temp_cs.add(var != value)
                    self._reset(temp_cs.to_string())
            return list(result)

    def _optimize_fancy(self, constraints: ConstraintSet, x: BitVec, goal: str, max_iter=10000):
        """
        Iteratively finds the maximum or minimum value for the operation
        (Normally Operators.UGT or Operators.ULT)

        :param constraints: constraints to take into account
        :param x: a symbol or expression
        :param goal: goal to achieve, either 'maximize' or 'minimize'
        :param max_iter: maximum number of iterations allowed
        """
        # TODO: consider adding a mode to return best known value on timeout
        assert goal in ("maximize", "minimize")
        operation = {"maximize": Operators.UGE, "minimize": Operators.ULE}[goal]

        with constraints as temp_cs:
            X = temp_cs.new_bitvec(x.size)
            temp_cs.add(X == x)
            aux = temp_cs.new_bitvec(X.size, name="optimized_")
            self._reset(temp_cs.to_string())

            self._assert(operation(X, aux))
            self._smtlib.send("(%s %s)" % (goal, aux.name))
            self._smtlib.send("(check-sat)")
            _status = self._smtlib.recv()
            if _status == "sat":
                return self._getvalue(aux)
            raise SolverError("Optimize failed")

    def get_value(self, constraints: ConstraintSet, *expressions):
        """
        Ask the solver for one possible result of given expressions using
        given set of constraints.
        """
        values = []
        start = time.time()
        with constraints.related_to(*expressions) as temp_cs:
            for expression in expressions:
                if not issymbolic(expression):
                    values.append(expression)
                    continue
                assert isinstance(expression, (Bool, BitVec, Array))
                if isinstance(expression, Bool):
                    var = temp_cs.new_bool()
                elif isinstance(expression, BitVec):
                    var = temp_cs.new_bitvec(expression.size)
                elif isinstance(expression, Array):
                    var = []
                    result = []
                    for i in range(expression.index_max):
                        subvar = temp_cs.new_bitvec(expression.value_bits)
                        var.append(subvar)
                        temp_cs.add(subvar == simplify(expression[i]))
                    self._reset(temp_cs.to_string())
                    if not self._is_sat():
                        raise SolverError(
                            "Solver could not find a value for expression under current constraint set"
                        )

                    for i in range(expression.index_max):
                        result.append(self.__getvalue_bv(var[i].name))
                    values.append(bytes(result))
                    if time.time() - start > consts.timeout:
                        raise SolverError("Timeout")
                    continue

                temp_cs.add(var == expression)

                self._reset(temp_cs.to_string())

                if not self._is_sat():
                    raise SolverError(
                        "Solver could not find a value for expression under current constraint set"
                    )

                if isinstance(expression, Bool):
                    values.append(self.__getvalue_bool(var.name))
                if isinstance(expression, BitVec):
                    values.append(self.__getvalue_bv(var.name))
            if time.time() - start > consts.timeout:
                raise SolverError("Timeout")

        if len(expressions) == 1:
            return values[0]
        else:
            return values


class Z3Solver(SMTLIBSolver):
    def __init__(self):
        """
        Build a Z3 solver instance.
        This is implemented using an external z3 solver (via a subprocess).
        See https://github.com/Z3Prover/z3
        """
        command = f"{consts.z3_bin} -t:{consts.timeout * 1000} -memory:{consts.memory} -smt2 -in"

        init, support_minmax, support_reset, multiple_check = self.__autoconfig()
        super().__init__(
            command=command,
            init=init,
            value_fmt=16,
            support_minmax=support_minmax,
            support_reset=support_reset,
            multiple_check=multiple_check,
            support_pushpop=True,
            debug=False,
        )

    def __autoconfig(self):
        init = [
            # http://smtlib.cs.uiowa.edu/logics-all.shtml#QF_AUFBV
            # Closed quantifier-free formulas over the theory of bitvectors and bitvector arrays extended with
            # free sort and function symbols.
            "(set-logic QF_AUFBV)",
            # The declarations and definitions will be scoped
            "(set-option :global-decls false)",
        ]

        # To cache what get-info returned; can be directly set when writing tests
        self.version = self._solver_version()
        support_reset = True

        if self.version > Version(4, 8, 4):
            # sam.moelius: Option "tactic.solve_eqs.context_solve" was turned on by this commit in z3:
            #   https://github.com/Z3Prover/z3/commit/3e53b6f2dbbd09380cd11706cabbc7e14b0cc6a2
            # Turning it off greatly improves Manticore's performance on test_integer_overflow_storageinvariant
            # in test_consensys_benchmark.py.
            init.append("(set-option :tactic.solve_eqs.context_solve false)")

        support_minmax = self.version >= Version(4, 4, 1)

        # Certain version of Z3 fails to handle multiple check-sat
        # https://gist.github.com/feliam/0f125c00cb99ef05a6939a08c4578902
        multiple_check = self.version < Version(4, 8, 7)
        return init, support_minmax, support_reset, multiple_check

    def _solver_version(self) -> Version:
        """
        If we fail to parse the version, we assume z3's output has changed, meaning it's a newer
        version than what's used now, and therefore ok.

        Anticipated version_cmd_output format: 'Z3 version 4.4.2'
                                               'Z3 version 4.4.5 - 64 bit - build hashcode $Z3GITHASH'
        """
        try:
            received_version = check_output([f"{consts.z3_bin}", "--version"])
            Z3VERSION = re.compile(
                r".*(?P<major>([0-9]+))\.(?P<minor>([0-9]+))\.(?P<patch>([0-9]+)).*"
            )
            m = Z3VERSION.match(received_version.decode("utf-8"))
            major, minor, patch = map(
                int, (m.group("major"), m.group("minor"), m.group("patch"))  # type: ignore
            )
            parsed_version = Version(major, minor, patch)
        except (ValueError, TypeError) as e:
            logger.warning(
                f"Could not parse Z3 version: '{str(received_version)}'. Assuming compatibility."
            )
            parsed_version = Version(float("inf"), float("inf"), float("inf"))
        return parsed_version


class YicesSolver(SMTLIBSolver):
    def __init__(self):
        init = ["(set-logic QF_AUFBV)"]
        command = f"{consts.yices_bin} --timeout={consts.timeout}  --incremental"
        super().__init__(
            command=command,
            init=init,
            value_fmt=2,
            debug=False,
            support_minmax=False,
            support_reset=False,
        )


class CVC4Solver(SMTLIBSolver):
    def __init__(self):
        init = ["(set-logic QF_AUFBV)", "(set-option :produce-models true)"]
        command = f"{consts.cvc4_bin} --lang=smt2 --incremental"
        super().__init__(command=command, value_fmt=10, init=init)


class SelectedSolver:
    choice = None

    @classmethod
    def instance(cls):
        if consts.solver == consts.solver.auto:
            if cls.choice is None:
                if shutil.which(consts.yices_bin):
                    cls.choice = consts.solver.yices
                elif shutil.which(consts.z3_bin):
                    cls.choice = consts.solver.z3
                elif shutil.which(consts.cvc4_bin):
                    cls.choice = consts.solver.cvc4
                else:
                    raise SolverException(
                        f"No Solver not found. Install one ({consts.yices_bin}, {consts.z3_bin}, {consts.cvc4_bin})."
                    )
        else:
            cls.choice = consts.solver

        SelectedSolver = {"cvc4": CVC4Solver, "yices": YicesSolver, "z3": Z3Solver}[cls.choice.name]
        return SelectedSolver.instance()



from ..core.smtlib import SelectedSolver, ConstraintSet, Operators, issymbolic, BitVec


def compare_sockets(cs, socket1, socket2):
    """
    This method compares Socket objects for equality using the buffer and peer attributes.
    It uses `compare_buffers` for checking buffer attributes for equality.
    It calls itself for comparing peer Socket objects.
    Returns True if the Socket objects are equal, false otherwise.
    :param cs: ConstraintSet to be used for checking Socket.buffer for semantic equality using `SelectedSolver.instance().must_be_true()`
    :param socket1: one of two Socket objects to be compared for equality against socket2
    :param socket2: one of two Socket objects to be compared for equality against socket1
    :return: True, if the Socket objects are found to be equal, False otherwise
    """
    if socket1 is None:
        return socket2 is None
    if socket2 is None:
        return socket1 is None
    if not compare_buffers(cs, socket1.buffer, socket2.buffer):
        return False
    return compare_sockets(cs, socket1.peer, socket2.peer)


def compare_buffers(cs, buffer1, buffer2):
    """
    This method compares the two List objects for equality using the `SelectedSolver.instance().must_be_true()` call.
    :param cs: ConstraintSet to be used for checking buffer1 for semantic equality with buffer2 using `SelectedSolver.instance().must_be_true()`
    :param buffer1: one of two List objects to be compared for equality against buffer2
    :param buffer2: one of two List objects to be compared for equality against buffer1
    :return: True, if the List objects are equal, False otherwise
    """
    if len(buffer1) != len(buffer2):
        return False
    for b1, b2 in zip(buffer1, buffer2):
        cond = cs.migrate(b1 == b2)
        if not SelectedSolver.instance().must_be_true(cs, cond):
            return False
    return True


def merge_constraints(constraints1, constraints2):
    """
    :param constraints1: one of two ConstraintSet objects to be merged
    :param constraints2: second of two ConstraintSet objects to be merged
    :return: (Expression, Expression, ConstraintSet) where the first and second Expression objects are conjunctions of
    of all the constraints in constraints1 and constraints2 respectively. The ConstraintSet is an object that contains
    a single constraint that is a logical OR of these two Expression objects.
    """
    exp1 = constraints1.constraints[0]
    for i in range(1, len(constraints1.constraints)):
        exp1 = exp1 & constraints1.constraints[i]
    exp2 = constraints2.constraints[0]
    for i in range(1, len(constraints2.constraints)):
        exp2 = exp2 & constraints2.constraints[i]
    merged_constraint = ConstraintSet()
    merged_constraint.add(exp1 | exp2)
    return exp1, exp2, merged_constraint


def compare_byte_vals(mem1, mem2, addr, merged_constraint):
    """
    Compares values in memory at address `addr`, returns True if they are semantically equal, False otherwise
    :param mem1: first of two memory objects we want to use for comparison
    :param mem2: second of two memory objects we want to use for comparison
    :param addr: address at which bytes values are to be compared
    :param merged_constraint: ConstraintSet to be used when using the call to `SelectedSolver.instance().must_be_true()`
    :return: returns True if 1 byte values at address `addr` in `mem1` and `mem2` are semantically equal, False otherwise
    """
    val1 = mem1.read(addr, 1)
    val2 = mem2.read(addr, 1)
    # since we only read a single byte value, these lists should only have one entry in them
    assert len(val1) == 1 and len(val2) == 1
    cond_to_check = merged_constraint.migrate(val1[0] == val2[0])
    if not SelectedSolver.instance().must_be_true(merged_constraint, cond_to_check):
        return False
    else:
        return True


# TODO move this comparison into an Executor API that uses an internal State API
def compare_mem(mem1, mem2, merged_constraint):
    """
    This method compares the number of maps, and then their names, permissions, start, and end values.
    If they all match, then it compares the concrete byte values for equality.
    If those match too, it then compares _symbols attribute values for equality if the two memory objects are of
    type SMemory.
    :param mem1: one of two memory objects to be compared
    :param mem2: second of two memory objects to be compared
    :param merged_constraint: ConstraintSet object that is to be used with `SelectedSolver.instance().must_be_true()` calls to check the
    memory objects for semantic equality
    :return: True, if the memory objects are equal, False otherwise
    """
    maps1 = sorted(list(mem1.maps))
    maps2 = sorted(list(mem2.maps))
    if len(maps1) != len(maps2):
        return False
    ret_val = None
    for m1, m2 in zip(maps1, maps2):
        if m1 != m2:  # compares the maps' names, permissions, starts, and ends
            ret_val = False
            break
        # Compare concrete byte values in the data in these memory maps for equality
        bytes1 = m1[m1.start : m1.end]
        bytes2 = m2[m2.start : m2.end]
        if bytes1 != bytes2:
            ret_val = False
            break
    checked_addrs = []
    # compare symbolic byte values in memory
    # hack to avoid importing SMemory because that import introduces a circular dependency on ManticoreBase
    if mem1.__class__.__name__ == "SMemory" and ret_val is not None:
        for addr1, _ in mem1._symbols.items():
            checked_addrs.append(addr1)
            if not compare_byte_vals(mem1, mem2, addr1, merged_constraint):
                ret_val = False
                break
    # hack to avoid importing SMemory because that import introduces a circular dependency on ManticoreBase
    if mem2.__class__.__name__ == "SMemory" and ret_val is not None:
        for addr2, _ in mem2._symbols.items():
            if addr2 not in checked_addrs:
                if not compare_byte_vals(mem1, mem2, addr2, merged_constraint):
                    ret_val = False
                    break
    if ret_val is not None:
        return ret_val
    else:
        return True


def is_merge_possible(state1, state2, merged_constraint):
    """
    Checks if a merge is possible by checking Input, Output sockets, symbolic_files, syscall_trace, and memory
    for equality.
    :param state1: one of two possible states we want to check for mergeability
    :param state2: second of two possible states we want to check for mergeability
    :param merged_constraint: ConstraintSet of merged constraints for state1 and state2
    :return: True, if state1 and state2 can be merged, False if otherwise
    """
    platform1 = state1.platform
    platform2 = state2.platform

    ret_val = None

    # compare input and output sockets of the states
    if not compare_sockets(
        merged_constraint, platform1.input, platform2.input
    ) or not compare_sockets(merged_constraint, platform1.output, platform2.output):
        ret_val = False, "inequivalent socket operations"

    # compare symbolic files opened by the two states
    if ret_val is None and platform1.symbolic_files != platform2.symbolic_files:
        ret_val = False, "inequivalent symbolic files"

    # compare system call traces of the two states
    if ret_val is None and len(platform1.syscall_trace) != len(platform2.syscall_trace):
        ret_val = False, "inequivalent syscall trace lengths"
    if ret_val is None:
        for i, (name1, fd1, data1) in enumerate(platform1.syscall_trace):
            (name2, fd2, data2) = platform2.syscall_trace[i]
            if not (
                name1 == name2 and fd1 == fd2 and compare_buffers(merged_constraint, data1, data2)
            ):
                ret_val = False, "inequivalent syscall traces"
                break

    # compare memory of the two states
    if ret_val is None and not compare_mem(state1.mem, state2.mem, merged_constraint):
        ret_val = False, "inequivalent memory"
    if ret_val is not None:
        return ret_val
    else:
        return True, None


def merge_cpu(cpu1, cpu2, state, exp1, merged_constraint):
    """
    Merge CPU objects into the state.CPU
    :param cpu1: one of two CPU objects that we wish to merge
    :param cpu2: second of two CPU objects that we wish to merge
    :param state: the state whose CPU attribute we will be updating
    :param exp1: the expression that if satisfiable will cause the CPU registers to take corresponding values from
    `cpu1`, else they will take corresponding values from `cpu2`
    :param merged_constraint: ConstraintSet under which we would want inequality between CPU register values to be
    satisfiable as checked using `SelectedSolver.instance().must_be_true()`
    :return: List of registers that were merged
    """
    merged_regs = []
    for reg in cpu1.canonical_registers:
        val1 = cpu1.read_register(reg)
        val2 = cpu2.read_register(reg)
        if isinstance(val1, BitVec) and isinstance(val2, BitVec):
            assert val1.size == val2.size
        if issymbolic(val1) or issymbolic(val2) or val1 != val2:
            val1_migrated = merged_constraint.migrate(val1)
            val2_migrated = merged_constraint.migrate(val2)
            if SelectedSolver.instance().must_be_true(
                merged_constraint, val1_migrated != val2_migrated
            ):
                merged_regs.append(reg)
                if cpu1.regfile.sizeof(reg) == 1:
                    state.cpu.write_register(reg, Operators.ITE(exp1, val1_migrated, val2_migrated))
                else:
                    state.cpu.write_register(
                        reg,
                        Operators.ITEBV(
                            cpu1.regfile.sizeof(reg), exp1, val1_migrated, val2_migrated
                        ),
                    )
    return merged_regs


def merge(state1, state2, exp1, merged_constraint):
    """
    Merge state1 and state2 into a single state
    :param state1:
    :param state2:
    :param exp1:
    :param merged_constraint:
    :return: the state that is the result of the merging of `state1` and `state2`
    """
    merged_state = state1
    merged_reg_list = merge_cpu(state1.cpu, state2.cpu, merged_state, exp1, merged_constraint)
    print("Merged registers: ")
    print(*merged_reg_list, sep=",")
    merged_state.constraints = merged_constraint
    return merged_state



from .expression import (
    BitVec,
    BitVecExtract,
    BitVecSignExtend,
    BitVecZeroExtend,
    BitVecConstant,
    BitVecConcat,
    Bool,
    BitVecITE,
    BoolConstant,
    BoolITE,
)
from . import issymbolic
import math


def ORD(s):
    if isinstance(s, BitVec):
        if s.size == 8:
            return s
        else:
            return BitVecExtract(s, 0, 8)
    elif isinstance(s, int):
        return s & 0xFF
    else:
        return ord(s)


def CHR(s):
    if isinstance(s, BitVec):
        if s.size == 8:
            return s
        else:
            return BitVecExtract(s, 0, 8)
    elif isinstance(s, int):
        return bytes([s & 0xFF])
    else:
        assert len(s) == 1
        return s


def NOT(a):
    if isinstance(a, bool):
        return not a
    if isinstance(a, (Bool, int)):
        return ~a
    return a == False


def AND(a, b, *others):
    if len(others) > 0:
        b = AND(b, others[0], *others[1:])
    if isinstance(a, Bool):
        return a & b
    if isinstance(b, Bool):
        return b & a
    assert isinstance(a, bool) and isinstance(b, bool)
    return a & b


def OR(a, b, *others):
    if len(others) > 0:
        b = OR(b, others[0], *others[1:])
    if isinstance(a, Bool):
        return a | b
    if isinstance(b, Bool):
        return b | a
    result = a | b
    if isinstance(result, (BitVec, int)):
        result = ITE(result != 0, True, False)
    return result


def UGT(a, b):
    if isinstance(a, BitVec):
        return a.ugt(b)
    if isinstance(b, BitVec):
        return b.ult(a)
    if a < 0:
        a = a & 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    if b < 0:
        b = b & 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF

    return a > b


def UGE(a, b):
    if isinstance(a, BitVec):
        return a.uge(b)
    if isinstance(b, BitVec):
        return b.ule(a)
    if a < 0:
        a = a & 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    if b < 0:
        b = b & 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF

    return a >= b


def ULT(a, b):
    if isinstance(a, BitVec):
        return a.ult(b)
    if isinstance(b, BitVec):
        return b.ugt(a)
    if a < 0:
        a = a & 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    if b < 0:
        b = b & 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF

    return a < b


def ULE(a, b):
    if isinstance(a, BitVec):
        return a.ule(b)
    if isinstance(b, BitVec):
        return b.uge(a)
    if a < 0:
        a = a & 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    if b < 0:
        b = b & 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF

    return a <= b


def EXTRACT(x, offset, size):
    if isinstance(x, BitVec) and isinstance(offset, BitVec):
        return BitVecExtract(x >> offset, 0, size)
    elif isinstance(x, BitVec):
        if offset == 0 and size == x.size:
            return x
        return BitVecExtract(x, offset, size)
    else:
        return (x >> offset) & ((1 << size) - 1)


def SEXTEND(x, size_src, size_dest):
    if isinstance(x, int):
        if x >= (1 << (size_src - 1)):
            x -= 1 << size_src
        return x & ((1 << size_dest) - 1)
    assert x.size == size_src
    return BitVecSignExtend(x, size_dest)


def ZEXTEND(x, size):
    if isinstance(x, int):
        return x & ((1 << size) - 1)
    assert isinstance(x, BitVec) and size - x.size >= 0
    if size - x.size > 0:
        return BitVecZeroExtend(size, x)
    else:
        return x


def CONCAT(total_size, *args):
    arg_size = total_size // len(args)
    if any(issymbolic(x) for x in args):
        if len(args) > 1:

            def cast(x):
                if isinstance(x, int):
                    return BitVecConstant(arg_size, x)
                return x

            return BitVecConcat(total_size, *list(map(cast, args)))
        else:
            return args[0]
    else:
        result = 0
        for arg in args:
            result = (result << arg_size) | (arg & ((1 << arg_size) - 1))
        return result


def ITE(cond, true_value, false_value):
    assert isinstance(true_value, (Bool, bool, BitVec, int))
    assert isinstance(false_value, (Bool, bool, BitVec, int))
    assert isinstance(cond, (Bool, bool))
    if isinstance(cond, bool):
        if cond:
            return true_value
        else:
            return false_value

    if isinstance(true_value, bool):
        true_value = BoolConstant(true_value)

    if isinstance(false_value, bool):
        false_value = BoolConstant(false_value)

    return BoolITE(cond, true_value, false_value)


def ITEBV(size, cond, true_value, false_value):
    if isinstance(cond, BitVec):
        cond = cond.Bool()
    if isinstance(cond, int):
        cond = cond != 0

    assert isinstance(cond, (Bool, bool))
    assert isinstance(true_value, (BitVec, int))
    assert isinstance(false_value, (BitVec, int))
    assert isinstance(size, int)

    if isinstance(cond, BoolConstant) and not cond.taint:
        cond = cond.value

    if isinstance(cond, bool):
        if cond:
            return true_value
        else:
            return false_value

    if isinstance(true_value, int):
        true_value = BitVecConstant(size, true_value)

    if isinstance(false_value, int):
        false_value = BitVecConstant(size, false_value)
    return BitVecITE(size, cond, true_value, false_value)


def UDIV(dividend, divisor):
    if isinstance(dividend, BitVec):
        return dividend.udiv(divisor)
    elif isinstance(divisor, BitVec):
        return divisor.rudiv(dividend)
    assert dividend >= 0 or divisor > 0  # unsigned-es
    return dividend // divisor


def SDIV(a, b):
    if isinstance(a, BitVec):
        return a.sdiv(b)
    elif isinstance(b, BitVec):
        return b.rsdiv(a)
    return int(math.trunc(float(a) / float(b)))


def SMOD(a, b):
    if isinstance(a, BitVec):
        return a.smod(b)
    elif isinstance(b, BitVec):
        return b.rsmod(a)
    return int(math.fmod(a, b))


def SREM(a, b):
    if isinstance(a, BitVec):
        return a.srem(b)
    elif isinstance(b, BitVec):
        return b.rsrem(a)
    elif isinstance(a, int) and isinstance(b, int):
        return a - int(a / b) * b
    return a % b


def UREM(a, b):
    if isinstance(a, BitVec):
        return a.urem(b)
    elif isinstance(b, BitVec):
        return b.rurem(a)
    return a % b


def SAR(size, a, b):
    assert isinstance(size, int)
    if isinstance(b, BitVec) and b.size != size:
        b = ZEXTEND(b, size)
    if isinstance(a, BitVec):
        assert size == a.size
        return a.sar(b)
    elif isinstance(b, BitVec):
        return BitVecConstant(size, a).sar(b)
    else:
        tempDest = a
        tempCount = b
        sign = tempDest & (1 << (size - 1))
        while tempCount != 0:
            tempDest = (tempDest >> 1) | sign
            tempCount = tempCount - 1
        return tempDest


def ABS(a):
    if issymbolic(a):
        return ITEBV(a.size, a < 0, -a, a)
    else:
        return abs(a)



from abc import ABCMeta, abstractmethod
from weakref import WeakValueDictionary
from dataclasses import dataclass, field
from ..core.smtlib import (
    Operators,
    ConstraintSet,
    arithmetic_simplify,
    SelectedSolver,
    TooManySolutions,
    BitVec,
    BitVecConstant,
    expression,
    issymbolic,
    Expression,
)
from ..native.mappings import mmap, munmap
from ..utils.helpers import interval_intersection
from ..utils import config

import functools
import logging

from typing import Dict, Generator, Iterable, List, MutableMapping, Optional, Set, Union

logger = logging.getLogger(__name__)

consts = config.get_group("native")
consts.add(
    "fast_crash",
    default=False,
    description="If True, throws a memory safety error if ANY concretization of a pointer is"
    " out of bounds. Otherwise, forks into valid and invalid memory access states.",
)
solver = SelectedSolver.instance()


class MemoryException(Exception):
    """
    Memory exceptions
    """

    def __init__(self, message: str, address=None):
        """
        Builds a memory exception.

        :param message: exception message.
        :param address: memory address where the exception occurred.
        """
        self.address = address
        self.message = message
        if address is not None and not issymbolic(address):
            self.message += f" <{address:x}>"

    def __str__(self):
        return self.message


class ConcretizeMemory(MemoryException):
    """
    Raised when a symbolic memory cell needs to be concretized.
    """

    def __init__(
        self,
        mem: "Memory",
        address: Union[int, Expression],
        size: int,
        message: Optional[str] = None,
        policy: str = "MINMAX",
    ):
        if message is None:
            self.message = f"Concretizing memory address {address} size {size}"
        else:
            self.message = message
        super().__init__(self.message, address)
        self.mem = mem
        self.address = address
        self.size = size
        self.policy = policy


class InvalidMemoryAccess(MemoryException):
    _message = "Invalid memory access"

    def __init__(self, address, mode: str):
        assert mode in "rwx"
        message = f"{self._message} (mode:{mode})"
        super(InvalidMemoryAccess, self).__init__(message, address)
        self.mode = mode


class InvalidSymbolicMemoryAccess(InvalidMemoryAccess):
    _message = "Invalid symbolic memory access"

    def __init__(self, address, mode: str, size, constraint):
        super(InvalidSymbolicMemoryAccess, self).__init__(address, mode)
        # the crashing constraint you need to assert
        self.constraint = constraint
        self.size = size


def _normalize(c):
    """
    Convert a byte-like value into a canonical byte (a value of type 'bytes' of len 1)

    :param c:
    :return:
    """
    if isinstance(c, int):
        return bytes([c])
    elif isinstance(c, str):
        return bytes([ord(c)])
    else:
        return c


class Map(object, metaclass=ABCMeta):
    """
    A memory map.

    It represents a convex chunk of memory with a start and an end address.
    It may be implemented as an actual file mapping or as a StringIO/bytearray.

    >>>           ######################################
                  ^                                    ^
                start                                 end

    """

    def __init__(self, start: int, size: int, perms: str, name=None):
        """
        Abstract memory map.

        :param start: the first valid address.
        :param size: the size of the map.
        :param perms: the access permissions of the map (rwx).
        """
        assert isinstance(start, int) and start >= 0, "Invalid start address"
        assert isinstance(size, int) and size > 0, "Invalid end address"

        super().__init__()
        self._start = start
        self._end = start + size
        self._set_perms(perms)
        self._name = name

    def _get_perms(self) -> str:
        """ Gets the access permissions of the map. """
        return self._perms

    def _set_perms(self, perms: str) -> None:
        """
        Sets the access permissions of the map.

        :param perms: the new permissions.
        """
        assert (
            isinstance(perms, str)
            and len(perms) <= 3
            and perms.strip() in ["", "r", "w", "x", "rw", "r x", "rx", "rwx", "wx"]
        )
        self._perms = perms

    # Property
    perms = property(_get_perms, _set_perms)

    def access_ok(self, access) -> bool:
        """ Check if there is enough permissions for access """
        for c in access:
            if c not in self.perms:
                return False
        return True

    @property
    def start(self) -> int:
        return self._start

    @property
    def end(self) -> int:
        return self._end

    @property
    def name(self):
        return self._name

    def __len__(self):
        """Returns the current size in bytes. """
        return self._end - self._start

    def __repr__(self):
        """
        Returns the string representation of the map mapping.

        :rtype: str
        """
        return f"<{self.__class__.__name__} 0x{self.start:016x}-0x{self.end:016x} {self.perms}>"

    def __iter__(self):
        """
        Iterate all valid addresses
        """
        return iter(range(self._start, self._end))

    def __eq__(self, other) -> bool:
        return (
            self.start == other.start
            and self.end == other.end
            and self.perms == other.perms
            and self.name == other.name
        )

    def __lt__(self, other):
        if self.start != other.start:
            return self.start < other.start
        if self.end != other.end:
            return self.end < other.end
        # go by each char permission
        if self.perms != other.perms:
            return self.perms < other.perms
        return self.name < other.name

    def __hash__(self) -> int:
        return object.__hash__(self)

    def _in_range(self, index) -> bool:
        """ Returns True if index is in range """
        if isinstance(index, slice):
            in_range = (
                index.start < index.stop and index.start >= self.start and index.stop <= self.end
            )
        else:
            in_range = index >= self.start and index <= self.end
        return in_range

    def _get_offset(self, index):
        """
        Translates the index to the internal offsets.

        self.start   -> 0
        self.start+1 -> 1
        ...
        self.end     -> len(self)
        """
        if not self._in_range(index):
            raise IndexError("Map index out of range")
        if isinstance(index, slice):
            index = slice(index.start - self.start, index.stop - self.start)
        else:
            index -= self.start
        return index

    @abstractmethod
    def __getitem__(self, index):
        """
        Reads a byte from an address or a sequence of bytes from a range of addresses

        :param index: the address or slice where to obtain the bytes from.
        :return: the character or sequence at the specified address.
        :rtype: byte or array
        """

    @abstractmethod
    def __setitem__(self, index, value):
        """
        Writes a byte to an address or a sequence of bytes to a range of addresses

        :param index: the address or slice where to put the data.
        :param value: byte or sequence of bytes to put in this map.
        """

    @abstractmethod
    def split(self, address):
        """
        Split the current map into two mappings

        :param address: The address at which to split the Map.
        """


class AnonMap(Map):
    """ A concrete anonymous memory map """

    def __init__(self, start: int, size: int, perms: str, data_init=None, name=None):
        """
        Builds a concrete anonymous memory map.

        :param start: the first valid address of the map.
        :param size: the size of the map.
        :param perms: the access permissions of the map.
        :param data_init: the data to initialize the map.
        """
        super().__init__(start, size, perms, name)
        self._data = bytearray(size)
        if data_init is not None:
            assert len(data_init) <= size, "More initial data than reserved memory"
            # check that the values this slice points to are ints
            if isinstance(data_init[0], int):
                self._data[0 : len(data_init)] = data_init
            else:
                self._data[0 : len(data_init)] = [ord(s) for s in data_init]

    def __reduce__(self):
        return (self.__class__, (self.start, len(self), self.perms, self._data, self.name))

    def split(self, address):
        if address <= self.start:
            return None, self
        if address >= self.end:
            return self, None

        assert address > self.start and address < self.end
        head = AnonMap(self.start, address - self.start, self.perms, self[self.start : address])
        tail = AnonMap(address, self.end - address, self.perms, self[address : self.end])
        return head, tail

    def __setitem__(self, index, value):
        assert not isinstance(index, slice) or len(value) == index.stop - index.start
        index = self._get_offset(index)

        if issymbolic(value[0]) and isinstance(self._data, bytearray):
            self._data = [
                Operators.ORD(b) for b in self._data
            ]  # completely convert everything to a list if we need to store some symbolic data, just copying fpse

        if isinstance(index, slice):
            if not isinstance(value[0], int):
                value = [Operators.ORD(n) for n in value]
            self._data[index] = value
        else:
            self._data[index] = Operators.ORD(value)

    def __getitem__(self, index):
        index = self._get_offset(index)
        if isinstance(index, slice):
            return [Operators.CHR(i) for i in self._data[index]]
        return Operators.CHR(self._data[index])


class ArrayMap(Map):
    def __init__(
        self, start: int, size: int, perms: str, index_bits, backing_array=None, name=None
    ):
        super().__init__(start, size, perms, name)
        if name is None:
            name = "ArrayMap_{:x}".format(start)
        if backing_array is not None:
            self._array = backing_array
        else:
            self._array = expression.ArrayProxy(
                expression.ArrayVariable(index_bits, index_max=size, value_bits=8, name=name)
            )

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.start,
                len(self),
                self._perms,
                self._array.index_bits,
                self._array,
                self._array.name,
            ),
        )

    def __setitem__(self, key, value):
        self._array[key] = value

    def __getitem__(self, key):
        return self._array[key]

    def split(self, address: int):
        if address <= self.start:
            return None, self
        if address >= self.end:
            return self, None

        assert self.start < address < self.end
        index_bits, value_bits = self._array.index_bits, self._array.value_bits

        left_size, right_size = address - self.start, self.end - address
        left_name, right_name = ["{}_{:d}".format(self._array.name, i) for i in range(2)]

        head_arr = expression.ArrayProxy(
            expression.ArrayVariable(index_bits, left_size, value_bits, name=left_name)
        )
        tail_arr = expression.ArrayProxy(
            expression.ArrayVariable(index_bits, right_size, value_bits, name=right_name)
        )

        head = ArrayMap(self.start, left_size, self.perms, index_bits, head_arr, left_name)
        tail = ArrayMap(address, right_size, self.perms, index_bits, tail_arr, right_name)

        return head, tail


class FileMap(Map):
    """
    A file map.

    A  file is mapped in multiples of the page size.  For a file that is not a
    multiple of the page size, the remaining memory is zeroed when mapped, and
    writes to that region are not written out to the file. The effect of
    changing the size of the underlying file of a mapping on the pages that
    correspond to added or removed regions of the file is unspecified.
    """

    def __init__(
        self, addr: int, size: int, perms: str, filename: str, offset: int = 0, overlay=None
    ):
        """
        Builds a map of memory  initialized with the content of filename.

        :param addr: the first valid address of the file map.
        :param size: the size of the file map.
        :param perms: the access permissions of the file map.
        :param filename: the file to map in memory.
        :param offset: the offset into the file where to start the mapping. \
                This offset must be a multiple of pagebitsize.
        """
        super().__init__(addr, size, perms)
        assert isinstance(offset, int)
        assert offset >= 0
        self._filename = filename
        self._offset = offset
        with open(filename, "r") as fileobject:
            fileobject.seek(0, 2)
            file_size = fileobject.tell()
            self._mapped_size = min(size, file_size - offset)
            self._data = mmap(fileobject.fileno(), offset, self._mapped_size)
        if overlay is not None:
            self._overlay = dict(overlay)
        else:
            self._overlay = dict()

    def __reduce__(self):
        return (
            self.__class__,
            (self.start, len(self), self.perms, self._filename, self._offset, self._overlay),
        )

    def __del__(self):
        if hasattr(self, "_data"):
            munmap(self._data, self._mapped_size)

    def __repr__(self):
        return f"<{self.__class__.__name__} [{self._filename}+{self._offset:x}] 0x{self.start:016x}-0x{self.end:016x} {self.perms}>"

    def __setitem__(self, index, value):
        assert not isinstance(index, slice) or len(value) == index.stop - index.start
        index = self._get_offset(index)
        if isinstance(index, slice):
            for i in range(index.stop - index.start):
                self._overlay[index.start + i] = value[i]
        else:
            self._overlay[index] = value

    def __getitem__(self, index):
        def get_byte_at_offset(offset):
            if offset in self._overlay:
                return _normalize(self._overlay[offset])
            else:
                if offset >= self._mapped_size:
                    return b"\x00"  # , 'Extra data must initially be zero'
                return _normalize(self._data[offset])

        index = self._get_offset(index)
        if isinstance(index, slice):
            result = []
            for i in range(index.start, index.stop):
                result.append(get_byte_at_offset(i))
            return result
        else:
            return get_byte_at_offset(index)

    def split(self, address: int):
        if address <= self.start:
            return None, self
        if address >= self.end:
            return self, None

        assert self.start < address <= self.end
        head = COWMap(self, size=address - self.start)
        tail = COWMap(self, offset=address - self.start)
        return head, tail


class COWMap(Map):
    """
    Copy-on-write based map.
    """

    def __init__(
        self, parent: Map, offset: int = 0, perms: Optional[str] = None, size=None, **kwargs
    ):
        """
        A copy on write copy of parent. Writes to the parent after a copy on
        write are unspecified.

        :param parent: the parent map.
        :param offset: an offset within the parent map from where to create the new map.
        :param perms: Permissions on new mapping, or None if inheriting.
        :param size: the size of the new map or max.
        """
        assert isinstance(parent, Map)
        assert offset >= 0 and offset < len(parent)
        if size is None:
            size = len(parent) - offset
        assert parent.start + offset + size <= parent.end
        if perms is None:
            perms = parent.perms

        super().__init__(parent.start + offset, size, perms, **kwargs)
        self._parent = parent
        # See (https://github.com/python/mypy/issues/2427)
        self._parent.__setitem__ = False  # type: ignore
        self._cow: Dict = {}

    def __setitem__(self, index, value):
        assert self._in_range(index)
        if isinstance(index, slice):
            for i in range(index.stop - index.start):
                self._cow[index.start + i] = _normalize(value[i])
        else:
            self._cow[index] = _normalize(value)

    def __getitem__(self, index):
        assert self._in_range(index)

        if isinstance(index, slice):
            result = []
            for i in range(index.start, index.stop):
                c = self._cow.get(i, self._parent[i])
                result.append(_normalize(c))
            return result
        else:
            return _normalize(self._cow.get(index, self._parent[index]))

    def split(self, address: int):
        if address <= self.start:
            return None, self
        if address >= self.end:
            return self, None

        assert address > self.start and address < self.end
        head = COWMap(self, size=address - self.start)
        tail = COWMap(self, offset=address - self.start)
        return head, tail


class StubCPU:
    def _publish(self, *args, **kwargs):
        return None


@dataclass
class ProcSelfMapInfo:
    start: int
    end: int
    rwx_perms: str
    shared_perms: str = "-"
    offset: int = 0
    device: str = "00:00"
    inode: int = 0
    pathname: str = ""

    perms: str = field(init=False)

    def __post_init__(self):
        self.perms = self.rwx_perms.replace(" ", "-") + self.shared_perms
        if self.pathname == "stack":
            self.pathname = "[" + self.pathname + "]"

    def __str__(self):
        return f"{self.address} {self.perms:>4s} {self.offset:08x} {self.device} {self.inode:9} {self.pathname}"

    @property
    def address(self):
        return f"{self.start:016x}-{self.end:016x}"


class Memory(object, metaclass=ABCMeta):
    """
    The memory manager.
    This class handles all virtual memory mappings and symbolic chunks.
    """

    def __init__(self, maps: Optional[Iterable[Map]] = None, cpu=StubCPU()):
        """
        Builds a memory manager.
        """
        super().__init__()
        if maps is None:
            self._maps: Set[Map] = set()
        else:
            self._maps = set(maps)
        self.cpu = cpu
        self._page2map: MutableMapping[int, Map] = WeakValueDictionary()  # {page -> ref{MAP}}
        self._recording_stack: List = []
        for m in self._maps:
            for i in range(self._page(m.start), self._page(m.end)):
                assert i not in self._page2map
                self._page2map[i] = m

    def __reduce__(self):
        return (self.__class__, (self._maps, self.cpu))

    @property
    @abstractmethod
    def memory_bit_size(self) -> int:
        return 32

    @property
    @abstractmethod
    def page_bit_size(self) -> int:
        return 12

    @property
    def memory_size(self) -> int:
        return 1 << self.memory_bit_size

    @property
    def page_size(self) -> int:
        return 1 << self.page_bit_size

    @property
    def memory_mask(self) -> int:
        return self.memory_size - 1

    @property
    def page_mask(self) -> int:
        return self.page_size - 1

    @property
    def maps(self) -> Set[Map]:
        return self._maps

    def _ceil(self, address) -> int:
        """
        Returns the smallest page boundary value not less than the address.
        :param address: the address to calculate its ceil.
        :return: the ceil of C{address}.
        """
        return (((address - 1) + self.page_size) & ~self.page_mask) & self.memory_mask

    def _floor(self, address) -> int:
        """
        Returns largest page boundary value not greater than the address.

        :param address: the address to calculate its floor.
        :return: the floor of C{address}.
        """
        return address & ~self.page_mask

    def _page(self, address) -> int:
        """
        Calculates the page number of an address.

        :param address: the address to calculate its page number.
        :return: the page number of address.
        """
        return address >> self.page_bit_size

    def _search(self, size, start=None, counter=0) -> int:
        """
        Recursively searches the address space for enough free space to allocate C{size} bytes.

        :param size: the size in bytes to allocate.
        :param start: an address from where to start the search.
        :param counter: internal parameter to know if all the memory was already scanned.
        :return: the address of an available space to map C{size} bytes.
        :raises MemoryException: if there is no space available to allocate the desired memory.


        todo: Document what happens when you try to allocate something that goes round the address 32/64 bit representation.
        """
        assert size & self.page_mask == 0
        if start is None:
            end = {32: 0xF8000000, 64: 0x0000800000000000}[self.memory_bit_size]
            start = end - size
        else:
            if start > self.memory_size - size:
                start = self.memory_size - size
            end = start + size

        consecutive_free = 0
        for p in range(self._page(end - 1), -1, -1):
            if p not in self._page2map:
                consecutive_free += 0x1000
            else:
                consecutive_free = 0
            if consecutive_free >= size:
                return p << self.page_bit_size
            counter += 1
            if counter >= self.memory_size // self.page_size:
                raise MemoryException("Not enough memory")

        return self._search(size, self.memory_size - size, counter)

    def mmapFile(self, addr, size, perms, filename, offset=0):
        """
        Creates a new file mapping in the memory address space.

        :param addr: the starting address (took as hint). If C{addr} is C{0} the first big enough
                     chunk of memory will be selected as starting address.
        :param size: the contents of a file mapping are initialized using C{size} bytes starting
                     at offset C{offset} in the file C{filename}.
        :param perms: the access permissions to this memory.
        :param filename: the pathname to the file to map.
        :param offset: the contents of a file mapping are initialized using C{size} bytes starting
                      at offset C{offset} in the file C{filename}.
        :return: the starting address where the file was mapped.
        :rtype: int
        :raises error:
                   - 'Address shall be concrete' if C{addr} is not an integer number.
                   - 'Address too big' if C{addr} goes beyond the limit of the memory.
                   - 'Map already used' if the piece of memory starting in C{addr} and with length C{size} isn't free.
        """
        # If addr is NULL, the system determines where to allocate the region.
        assert addr is None or isinstance(addr, int), "Address shall be concrete"
        assert size > 0

        self.cpu._publish("will_map_memory", addr, size, perms, filename, offset)

        # address is rounded down to the nearest multiple of the allocation granularity
        if addr is not None:
            assert addr < self.memory_size, "Address too big"
            addr = self._floor(addr)

        # size value is rounded up to the next page boundary
        size = self._ceil(size)

        # If zero search for a spot
        addr = self._search(size, addr)

        # It should not be allocated
        for i in range(self._page(addr), self._page(addr + size)):
            assert i not in self._page2map, "Map already used"

        # Create the map
        m = FileMap(addr, size, perms, filename, offset)

        # Okay, ready to alloc
        self._add(m)

        logger.debug("New file-memory map @%x size:%x", addr, size)
        self.cpu._publish("did_map_memory", addr, size, perms, filename, offset, addr)
        return addr

    def mmap(self, addr, size, perms, data_init=None, name=None):
        """
        Creates a new mapping in the memory address space.

        :param addr: the starting address (took as hint). If C{addr} is C{0} the first big enough
                     chunk of memory will be selected as starting address.
        :param size: the length of the mapping.
        :param perms: the access permissions to this memory.
        :param data_init: optional data to initialize this memory.
        :param name: optional name to give to this mapping
        :return: the starting address where the memory was mapped.
        :raises error:
                   - 'Address shall be concrete' if C{addr} is not an integer number.
                   - 'Address too big' if C{addr} goes beyond the limit of the memory.
                   - 'Map already used' if the piece of memory starting in C{addr} and with length C{size} isn't free.
        :rtype: int

        """
        # If addr is NULL, the system determines where to allocate the region.
        assert addr is None or isinstance(addr, int), "Address shall be concrete"

        self.cpu._publish("will_map_memory", addr, size, perms, None, None)

        # address is rounded down to the nearest multiple of the allocation granularity
        if addr is not None:
            assert addr < self.memory_size, "Address too big"
            addr = self._floor(addr)

        # size value is rounded up to the next page boundary
        size = self._ceil(size)

        # If zero search for a spot
        addr = self._search(size, addr)

        # It should not be allocated
        for i in range(self._page(addr), self._page(addr + size)):
            assert i not in self._page2map, "Map already used"

        # Create the anonymous map
        m = AnonMap(start=addr, size=size, perms=perms, data_init=data_init, name=name)

        # Okay, ready to alloc
        self._add(m)

        logger.debug("New memory map @%x size:%x", addr, size)

        self.cpu._publish("did_map_memory", addr, size, perms, None, None, addr)
        return addr

    def _add(self, m: Map) -> None:
        assert isinstance(m, Map)
        assert m not in self._maps
        assert m.start & self.page_mask == 0
        assert m.end & self.page_mask == 0
        self._maps.add(m)
        # updating the page to map translation
        for i in range(self._page(m.start), self._page(m.end)):
            self._page2map[i] = m

    def _del(self, m: Map) -> None:
        assert isinstance(m, Map)
        assert m in self._maps
        # remove m pages from the page2maps..
        for p in range(self._page(m.start), self._page(m.end)):
            del self._page2map[p]
        # remove m from the maps set
        self._maps.remove(m)

    def map_containing(self, address: int) -> Map:
        """
        Returns the L{MMap} object containing the address.

        :param address: the address to obtain its mapping.
        :rtype: L{MMap}

        @todo: symbolic address
        """
        page_offset = self._page(address)
        if page_offset not in self._page2map:
            raise MemoryException("Page not mapped", address)
        return self._page2map[page_offset]

    def mappings(self):
        """
        Returns a sorted list of all the mappings for this memory.

        :return: a list of mappings.
        :rtype: list
        """
        result = []
        for m in self.maps:
            if isinstance(m, AnonMap):
                result.append((m.start, m.end, m.perms, 0, ""))
            elif isinstance(m, FileMap):
                result.append((m.start, m.end, m.perms, m._offset, m._filename))
            else:
                result.append((m.start, m.end, m.perms, 0, m.name))

        return sorted(result)

    def __str__(self) -> str:
        return "\n".join(
            [
                f'{start:016x}-{end:016x} {p:>4s} {offset:08x} {name or ""}'
                for start, end, p, offset, name in self.mappings()
            ]
        )

    def proc_self_mappings(self) -> List[ProcSelfMapInfo]:
        """
        Returns a sorted list of all the mappings for this memory for /proc/self/maps.
        Device, inode, and private/shared permissions are unsupported.
        Stack is the only memory section supported in the memory map (heap, vdso, etc.)
        are unsupported.
        Pathname is substituted by filename

        :return: a list of mappings.
        """
        result = []
        # TODO: Device, inode, and private/shared permissions are unsupported
        # TODO: Add complete paths
        for m in self.maps:
            if isinstance(m, AnonMap):
                if m.name is not None:
                    result.append(
                        ProcSelfMapInfo(m.start, m.end, rwx_perms=m.perms, pathname=m.name)
                    )
                else:
                    result.append(ProcSelfMapInfo(m.start, m.end, rwx_perms=m.perms))
            elif isinstance(m, FileMap):
                result.append(
                    ProcSelfMapInfo(
                        m.start, m.end, rwx_perms=m.perms, offset=m._offset, pathname=m._filename
                    )
                )
            else:
                result.append(ProcSelfMapInfo(m.start, m.end, rwx_perms=m.perms, pathname=m.name))
        return sorted(result, key=lambda m: m.start)

    @property
    def __proc_self__(self):
        self.proc_self_mappings()
        return "\n".join([f"{m}" for m in self.proc_self_mappings()])

    def _maps_in_range(self, start: int, end: int) -> Generator[Map, None, None]:
        """
        Generates the list of maps that overlaps with the range [start:end]
        """

        # Search for the first matching map
        addr = start
        while addr < end:
            if addr not in self:
                addr += self.page_size
            else:
                m = self._page2map[self._page(addr)]
                yield m
                addr = m.end

    def munmap(self, start, size):
        """
        Deletes the mappings for the specified address range and causes further
        references to addresses within the range to generate invalid memory
        references.

        :param start: the starting address to delete.
        :param size: the length of the unmapping.
        """
        start = self._floor(start)
        end = self._ceil(start + size)

        self.cpu._publish("will_unmap_memory", start, size)

        for m in self._maps_in_range(start, end):
            self._del(m)
            head, tail = m.split(start)
            middle, tail = tail.split(end)
            assert middle is not None
            if head:
                self._add(head)
            if tail:
                self._add(tail)

        self.cpu._publish("did_unmap_memory", start, size)
        logger.debug(f"Unmap memory @{start:x} size:{size:x}")

    def mprotect(self, start, size, perms):
        assert size > 0
        start = self._floor(start)
        end = self._ceil(start + size)

        self.cpu._publish("will_protect_memory", start, size, perms)

        for m in self._maps_in_range(start, end):
            self._del(m)

            head, tail = m.split(start)
            middle, tail = tail.split(end)

            assert middle is not None
            middle.perms = perms
            self._add(middle)

            if head:
                self._add(head)
            if tail:
                self._add(tail)

        self.cpu._publish("did_protect_memory", start, size, perms)

    # Permissions
    def __contains__(self, address):
        return self._page(address) in self._page2map

    def perms(self, index):
        # not happy with this interface.
        if isinstance(index, slice):
            # get the more restrictive set of perms for the range
            raise NotImplementedError("No perms for slices")
        else:
            return self.map_containing(index).perms

    def max_exec_size(self, addr, max_size):
        """
        Finds maximum executable memory size
        starting from addr and up to max_size.

        :param addr: the starting address.
        :param size: the maximum size.
        :param access: the wished access.
        """
        size = 0
        max_addr = addr + max_size
        while addr < max_addr:
            if addr not in self:
                return size
            m = self.map_containing(addr)
            # could it be ? if access in m.perm
            if not m.access_ok("x"):
                return size
            size += m.end - addr
            addr = m.end
        return max_size

    def access_ok(self, index, access, force=False):
        if isinstance(index, slice):
            assert index.stop - index.start >= 0
            addr = index.start
            while addr < index.stop:
                if addr not in self:
                    return False
                m = self.map_containing(addr)

                if not force and not m.access_ok(access):
                    return False

                until_next_page = min(m.end - addr, index.stop - addr)
                addr += until_next_page
            assert addr == index.stop
            return True
        else:
            if index not in self:
                return False
            m = self.map_containing(index)
            return force or m.access_ok(access)

    # write and read potentially symbolic bytes at symbolic indexes
    def read(self, addr, size, force: bool = False) -> List[bytes]:
        if not self.access_ok(slice(addr, addr + size), "r", force):
            raise InvalidMemoryAccess(addr, "r")

        assert size > 0
        result: List[bytes] = []
        stop = addr + size
        p = addr
        while p < stop:
            m = self.map_containing(p)

            _size = min(m.end - p, stop - p)
            result += m[p : p + _size]
            p += _size
        assert p == stop

        return result

    def push_record_writes(self):
        """
        Begin recording all writes. Retrieve all writes with `pop_record_writes()`
        """
        self._recording_stack.append([])

    def pop_record_writes(self):
        """
        Stop recording trace and return a `list[(address, value)]` of all the writes
        that occurred, where `value` is of type list[str]. Can be called without
        intermediate `pop_record_writes()`.

        For example::

            mem.push_record_writes()
                mem.write(1, 'a')
                mem.push_record_writes()
                    mem.write(2, 'b')
                mem.pop_record_writes()  # Will return [(2, 'b')]
            mem.pop_record_writes()  # Will return [(1, 'a'), (2, 'b')]

        Multiple writes to the same address will all be included in the trace in the
        same order they occurred.

        :return: list[tuple]
        """

        lst = self._recording_stack.pop()
        # Append the current list to a previously-started trace.
        if self._recording_stack:
            self._recording_stack[-1].extend(lst)
        return lst

    def write(self, addr, buf, force=False):
        size = len(buf)
        if not self.access_ok(slice(addr, addr + size), "w", force):
            raise InvalidMemoryAccess(addr, "w")
        assert size > 0
        stop = addr + size
        start = addr

        if self._recording_stack:
            self._recording_stack[-1].append((addr, buf))

        while addr < stop:
            m = self.map_containing(addr)
            size = min(m.end - addr, stop - addr)
            m[addr : addr + size] = buf[addr - start : addr - start + size]
            addr += size
        assert addr == stop

    def _get_size(self, size):
        return size

    def __setitem__(self, index, value):
        if isinstance(index, slice):
            size = self._get_size(index.stop - index.start)
            assert len(value) == size  # raise proper Error?
            self.write(index.start, value)
        else:
            self.write(index, (value,))

    def __getitem__(self, index):
        if isinstance(index, slice):
            result = self.read(index.start, index.stop - index.start)
        else:
            result = self.read(index, 1)[0]
        return result

    def __iter__(self):
        """
        Iterate all valid addresses
        """
        for page_addr in sorted(self._page2map.keys()):
            start = page_addr * self.page_size
            end = start + self.page_size
            for addr in range(start, end):
                yield addr


class SMemory(Memory):
    """
    The symbolic memory manager.
    This class handles all virtual memory mappings and symbolic chunks.

    :todo: improve comments
    """

    def __init__(self, constraints: ConstraintSet, symbols=None, *args, **kwargs):
        """
        Builds a memory.

        :param constraints:  a set of initial constraints
        :param symbols: Symbolic chunks in format: {chunk_start_addr: (condition, value), ...}

        `symbols` or `self._symbols` is a mapping of symbolic chunks/memory cells starting addresses
        to their condition and value.

        The condition of a symbolic chunk can be concrete (True/False) or symbolic. The value should
        always be symbolic (e.g. a BitVecVariable).
        """
        super().__init__(*args, **kwargs)
        assert isinstance(constraints, ConstraintSet)
        self._constraints = constraints
        self._symbols: Dict
        if symbols is None:
            self._symbols = {}
        else:
            self._symbols = dict(symbols)

    def __reduce__(self):
        return (self.__class__, (self.constraints, self._symbols, self._maps), {"cpu": self.cpu})

    @property
    def constraints(self):
        return self._constraints

    @constraints.setter
    def constraints(self, constraints):
        self._constraints = constraints

    def _get_size(self, size):
        if isinstance(size, BitVec):
            size = arithmetic_simplify(size)
        else:
            size = BitVecConstant(self.memory_bit_size, size)
        assert isinstance(size, BitVecConstant)
        return size.value

    def munmap(self, start, size):
        """
        Deletes the mappings for the specified address range and causes further
        references to addresses within the range to generate invalid memory
        references.

        :param start: the starting address to delete.
        :param size: the length of the unmapping.
        """
        for addr in range(start, start + size):
            if len(self._symbols) == 0:
                break
            if addr in self._symbols:
                del self._symbols[addr]
        super().munmap(start, size)

    def read(self, address, size, force=False):
        """
        Read a stream of potentially symbolic bytes from a potentially symbolic
        address

        :param address: Where to read from
        :param size: How many bytes
        :param force: Whether to ignore permissions
        :rtype: list
        """
        size = self._get_size(size)
        assert not issymbolic(size)
        if issymbolic(address):
            logger.debug(f"Reading {size} bytes from symbolic address {address}")
            try:
                solutions = self._try_get_solutions(address, size, "r", force=force)
                assert len(solutions) > 0
            except TooManySolutions as e:
                self.cpu._publish("will_solve", self.constraints, address, "minmax")
                m, M = solver.minmax(self.constraints, address)
                self.cpu._publish("did_solve", self.constraints, address, "minmax", (m, M))
                logger.debug(
                    f"Got TooManySolutions on a symbolic read. Range [{m:x}, {M:x}]. Not crashing!"
                )

                # The force param shouldn't affect this, as this is checking for unmapped reads, not bad perms
                crashing_condition = True
                for start, end, perms, offset, name in self.mappings():
                    if start <= M + size and end >= m:
                        if "r" in perms:
                            crashing_condition = Operators.AND(
                                Operators.OR((address + size).ult(start), address.uge(end)),
                                crashing_condition,
                            )

                self.cpu._publish("will_solve", self.constraints, crashing_condition, "can_be_true")
                can_crash = solver.can_be_true(self.constraints, crashing_condition)
                self.cpu._publish(
                    "did_solve", self.constraints, crashing_condition, "can_be_true", can_crash
                )
                if can_crash:
                    raise InvalidSymbolicMemoryAccess(address, "r", size, crashing_condition)

                # INCOMPLETE Result! We could also fork once for every map
                logger.info("INCOMPLETE Result! Using the sampled solutions we have as result")
                condition = False
                for base in e.solutions:
                    condition = Operators.OR(address == base, condition)
                from ..core.state import ForkState

                raise ForkState("Forking state on incomplete result", condition)

            # So here we have all potential solutions to address

            condition = False
            for base in solutions:
                condition = Operators.OR(address == base, condition)

            result = []
            # consider size ==1 to read following code
            for offset in range(size):
                # Given ALL solutions for the symbolic address
                for base in solutions:
                    addr_value = base + offset
                    byte = Operators.ORD(self.map_containing(addr_value)[addr_value])
                    if addr_value in self._symbols:
                        for condition, value in self._symbols[addr_value]:
                            byte = Operators.ITEBV(8, condition, Operators.ORD(value), byte)
                    if len(result) > offset:
                        result[offset] = Operators.ITEBV(8, address == base, byte, result[offset])
                    else:
                        result.append(byte)
                    assert len(result) == offset + 1
            return list(map(Operators.CHR, result))
        else:
            result = list(map(Operators.ORD, super().read(address, size, force)))
            for offset in range(size):
                if address + offset in self._symbols:
                    for condition, value in self._symbols[address + offset]:
                        if condition is True:
                            result[offset] = Operators.ORD(value)
                        else:
                            result[offset] = Operators.ITEBV(
                                8, condition, Operators.ORD(value), result[offset]
                            )
            return list(map(Operators.CHR, result))

    def write(self, address, value, force: bool = False) -> None:
        """
        Write a value at address.

        :param address: The address at which to write
        :type address: int or long or Expression
        :param value: Bytes to write
        :type value: str or list
        :param force: Whether to ignore permissions

        """
        size = len(value)
        if issymbolic(address):

            solutions = self._try_get_solutions(address, size, "w", force=force)

            for offset in range(size):
                for base in solutions:
                    condition = base == address
                    self._symbols.setdefault(base + offset, []).append((condition, value[offset]))
        else:

            # Symbolic writes are stored in self._symbols and concrete writes are offloaded to super().write(...)
            #
            # Symbolic values are written one by one and concrete are written in chunks
            # To do this, the logic below tracks start index of the part where concrete values are
            # and commits/triggers writes of those concrete values when a symbolic value is hit or after the loop
            #
            concrete_start = None

            for offset in range(size):
                ea = address + offset

                if issymbolic(value[offset]):
                    # Commit (offload) concrete write
                    if concrete_start is not None:
                        super().write(address + concrete_start, value[concrete_start:offset], force)
                        concrete_start = None

                    # Commit symbolic write if access is okay
                    if not self.access_ok(ea, "w", force):
                        raise InvalidMemoryAccess(ea, "w")
                    self._symbols[ea] = [(True, value[offset])]

                else:
                    # For concrete value, remove symbolic chunks that were stored there
                    # (Note: there might be many because of conditions later leading to different program paths)
                    if ea in self._symbols:
                        del self._symbols[ea]

                    # Set concrete part start index if previous values weren't concrete
                    if concrete_start is None:
                        concrete_start = offset

            # Commit (offload) the last concrete write
            if concrete_start is not None:
                super().write(address + concrete_start, value[concrete_start:], force)

    def _try_get_solutions(self, address, size, access, max_solutions=0x1000, force=False):
        """
        Try to solve for a symbolic address, checking permissions when reading/writing size bytes.

        :param Expression address: The address to solve for
        :param int size: How many bytes to check permissions for
        :param str access: 'r' or 'w'
        :param int max_solutions: Will raise if more solutions are found
        :param force: Whether to ignore permission failure
        :rtype: list
        """
        assert issymbolic(address)
        self.cpu._publish("will_solve", self.constraints, address, "get_all_values")
        solutions = solver.get_all_values(
            self.constraints, address, maxcnt=max_solutions, silent=True
        )
        self.cpu._publish("did_solve", self.constraints, address, "get_all_values", solutions)

        crashing_condition = False
        for base in solutions:
            if not self.access_ok(slice(base, base + size), access, force):
                crashing_condition = Operators.OR(address == base, crashing_condition)

        self.cpu._publish("will_solve", self.constraints, crashing_condition, "get_all_values")
        crash_or_not = solver.get_all_values(self.constraints, crashing_condition, maxcnt=3)
        self.cpu._publish(
            "did_solve", self.constraints, crashing_condition, "get_all_values", crash_or_not
        )

        if not consts.fast_crash and len(crash_or_not) == 2:
            from ..core.state import Concretize

            def setstate(state, _value):
                """ Roll back PC to redo last instruction """
                state.cpu.PC = state.cpu._last_pc

            raise Concretize(
                "Forking on memory safety", expression=crashing_condition, setstate=setstate
            )
        elif any(crash_or_not):
            raise InvalidSymbolicMemoryAccess(address, access, size, crashing_condition)

        return solutions


class LazySMemory(SMemory):
    """
    A fully symbolic memory.

    Currently does not support cross-page reads/writes.
    """

    def __init__(self, constraints, *args, **kwargs):
        super(LazySMemory, self).__init__(constraints, *args, **kwargs)
        self.backing_array = constraints.new_array(index_bits=self.memory_bit_size)
        self.backed_by_symbolic_store = set()

    def __reduce__(self):
        return (
            self.__class__,
            (self.constraints, self._symbols, self._maps),
            {
                "backing_array": self.backing_array,
                "backed_by_symbolic_store": self.backed_by_symbolic_store,
                "cpu": self.cpu,
            },
        )

    def __setstate__(self, state):
        self.backing_array = state["backing_array"]
        self.backed_by_symbolic_store = state["backed_by_symbolic_store"]

    def mmapFile(self, addr, size, perms, filename, offset=0):
        """
        Creates a new file mapping in the memory address space.

        :param addr: the starting address (took as hint). If C{addr} is C{0} the first big enough
                     chunk of memory will be selected as starting address.
        :param size: the contents of a file mapping are initialized using C{size} bytes starting
                     at offset C{offset} in the file C{filename}.
        :param perms: the access permissions to this memory.
        :param filename: the pathname to the file to map.
        :param offset: the contents of a file mapping are initialized using C{size} bytes starting
                      at offset C{offset} in the file C{filename}.
        :return: the starting address where the file was mapped.
        :rtype: int
        :raises error:
                   - 'Address shall be concrete' if C{addr} is not an integer number.
                   - 'Address too big' if C{addr} goes beyond the limit of the memory.
                   - 'Map already used' if the piece of memory starting in C{addr} and with length C{size} isn't free.
        """
        # If addr is NULL, the system determines where to allocate the region.
        assert addr is None or isinstance(addr, int), "Address shall be concrete"
        assert addr < self.memory_size, "Address too big"
        assert size > 0

        self.cpu._publish("will_map_memory", addr, size, perms, filename, offset)

        map = AnonMap(addr, size, perms)
        self._add(map)

        # address is rounded down to the nearest multiple of the allocation granularity
        if addr is not None:
            addr = self._floor(addr)

        # size value is rounded up to the next page boundary
        size = self._ceil(size)

        with open(filename, "rb") as f:
            fdata = f.read()  # fdata is a bytes now

        # this worked
        fdata = fdata[offset:]
        fdata = fdata.ljust(size, b"\0")

        for i in range(size):
            Memory.write(self, addr + i, chr(fdata[i]), force=True)

        logger.debug("New file-memory map @%x size:%x", addr, size)
        self.cpu._publish("did_map_memory", addr, size, perms, filename, offset, addr)

        return addr

    def _deref_can_succeed(self, mapping, address, size):
        if not issymbolic(address):
            return address >= mapping.start and address + size < mapping.end
        else:
            constraint = Operators.AND(address >= mapping.start, address + size < mapping.end)
            self.cpu._publish("will_solve", self.constraints, constraint, "can_be_true")
            deref_can_succeed = solver.can_be_true(self.constraints, constraint)
            self.cpu._publish(
                "did_solve", self.constraints, constraint, "can_be_true", deref_can_succeed
            )
            return deref_can_succeed

    def _import_concrete_memory(self, from_addr, to_addr):
        """
        for each address in this range need to read from concrete and write to symbolic
        it's possible that there will be invalid/unmapped addresses in this range. need to skip to next map if so
        also need to mark all of these addresses as now in the symbolic store

        :param int from_addr:
        :param int to_addr:
        :return:
        """
        logger.debug(
            "Importing concrete memory: {:x} - {:x} ({} bytes)".format(
                from_addr, to_addr, to_addr - from_addr
            )
        )

        for m in self.maps:
            span = interval_intersection(m.start, m.end, from_addr, to_addr)
            if span is None:
                continue

            start, stop = span

            for addr in range(start, stop):
                if addr in self.backed_by_symbolic_store:
                    continue

                self.backing_array[addr] = Memory.read(self, addr, 1)[0]
                self.backed_by_symbolic_store.add(addr)

    def _map_deref_expr(self, map, address):
        return Operators.AND(Operators.UGE(address, map.start), Operators.ULT(address, map.end))

    def _reachable_range(self, sym_address, size):
        self.cpu._publish("will_solve", self.constraints, sym_address, "minmax")
        addr_min, addr_max = solver.minmax(self.constraints, sym_address)
        self.cpu._publish(
            "did_solve", self.constraints, sym_address, "minmax", (addr_min, addr_max)
        )
        return addr_min, addr_max + size - 1

    def valid_ptr(self, address):
        assert issymbolic(address)

        expressions = [self._map_deref_expr(m, address) for m in self._maps]
        valid = functools.reduce(Operators.OR, expressions)

        return valid

    def invalid_ptr(self, address):
        return Operators.NOT(self.valid_ptr(address))

    def read(self, address, size, force=False):

        access_min, access_max = self._reachable_range(address, size)

        if issymbolic(address):
            self._import_concrete_memory(access_min, access_max)

        retvals = []
        # Not range() because `address` can be symbolic.
        addrs_to_access = [address + i for i in range(size)]
        for addr in addrs_to_access:

            # If the deref is symbolic, or it's backed by a symbolic store on a page
            # that is writeable, go from backing_array. (if page is r-- or r-x, it would
            # not have had symbolic data written to it.)
            if issymbolic(addr):
                from_array = True
            elif addr in self.backed_by_symbolic_store:
                m = self.map_containing(addr)
                from_array = not m or "w" in m.perms
            else:
                from_array = False

            if from_array:
                val = self.backing_array[addr]
            else:
                val = Memory.read(self, addr, 1)[0]

            retvals.append(val)

        return retvals

    def write(self, address, value, force=False):

        size = len(value)

        addrs_to_access = [address + i for i in range(size)]

        if issymbolic(address):
            access_min, access_max = self._reachable_range(address, size)

            self._import_concrete_memory(access_min, access_max)

            for addr, byte in zip(addrs_to_access, value):
                self.backing_array[addr] = Operators.ORD(byte)
        else:
            self.backed_by_symbolic_store -= set(addrs_to_access)
            Memory.write(self, address, value)

    def scan_mem(self, data_to_find):
        """
        Scan for concrete bytes in all mapped memory. Successively yield addresses of all matches.

        :param bytes data_to_find: String to locate
        :return:
        """

        # TODO: for the moment we just treat symbolic bytes as bytes that don't match.
        # for our simple test cases right now, the bytes we're interested in scanning
        # for will all just be there concretely
        # TODO: Can probably do something smarter here like Boyer-Moore, but unnecessary
        # if we're looking for short strings.

        # Querying mem with an index returns [bytes]
        if isinstance(data_to_find, bytes):
            data_to_find = [bytes([c]) for c in data_to_find]

        for mapping in sorted(self.maps):
            for ptr in mapping:
                if ptr + len(data_to_find) >= mapping.end:
                    break

                candidate = mapping[ptr : ptr + len(data_to_find)]

                # TODO: treat symbolic bytes as bytes that don't match. for our simple tests right now, the
                # bytes will be there concretely
                if issymbolic(candidate[0]):
                    break

                if candidate == data_to_find:
                    yield ptr


class Memory32(Memory):
    memory_bit_size: int = 32
    page_bit_size: int = 12


class Memory64(Memory):
    memory_bit_size: int = 64
    page_bit_size: int = 12


class SMemory32(SMemory):
    memory_bit_size: int = 32
    page_bit_size: int = 12


class SMemory32L(SMemory):
    memory_bit_size: int = 32
    page_bit_size: int = 13


class SMemory64(SMemory):
    memory_bit_size: int = 64
    page_bit_size: int = 12


class LazySMemory32(LazySMemory):
    memory_bit_size: int = 32
    page_bit_size: int = 12


class LazySMemory64(LazySMemory):
    memory_bit_size: int = 64
    page_bit_size: int = 12
